@@ -8,7 +8,8 @@
 
 
 ellipsoidhull <-
-    function(x, tol = 0.01, maxit = 5000, ret.wt = FALSE, ret.sqdist = FALSE)
+    function(x, tol = 0.01, maxit = 5000,
+             ret.wt = FALSE, ret.sqdist = FALSE, ret.pr = FALSE)
 {
     if(!is.matrix(x) || !is.numeric(x))
         stop("`x' must be numeric  n x p matrix")
@@ -20,36 +21,44 @@ ellipsoidhull <-
     if(n == 0) stop("no points without missing values")
     p <- ncol(x)
 
-    res <- .Fortran("spannel",
-                    n,
-                    ndep= p,
-                    dat = cbind(1., x),
-                    sqdist = double(n),
-                    l1 = double((p+1) ^ 2),
-                    double(p),
-                    double(p),
-                    prob = double(n),
-                    double(p+1),
-                    eps = as.double(tol),
-                    maxit = as.integer(maxit),
-                    ierr = as.integer(0),
-                    PACKAGE = "cluster")
+    res <- .C("spannel",
+              n,
+              ndep= p,
+              dat = cbind(1., x),
+              sqdist = double(n),
+              l1 = double((p+1) ^ 2),
+              double(p),
+              double(p),
+              prob = double(n),
+              double(p+1),
+              eps = as.double(tol),
+              maxit = as.integer(maxit),
+              ierr = integer(1),# 0 or non-zero
+              PACKAGE = "cluster")
     if(res$ierr != 0)
         cat("Error in Fortran routine computing the spanning ellipsoid,",
             "\n probably collinear data\n", sep="")
-    if(res$maxit == maxit)
-        warning("possibly not converged in", maxit, "iterations")
+    if(any(res$prob < 0) || all(res$prob == 0))
+        stop("computed some negative or all 0 `prob'abilities")
+    conv <- res$maxit  < maxit
+    if(!conv)
+        warning("possibly not converged in ", maxit, " iterations")
+    conv <- conv && res$ierr == 0
 
     cov <- cov.wt(x, res$prob)
     ## cov.wt() in R has extra wt[] scaling; revert here
     res <- list(loc = cov$center,
                 cov = cov$cov * (1 - sum(cov$wt^2)),
                 d2  = weighted.mean(res$sqdist, res$prob),
-                sqdist = if(ret.sqdist) res$sqdist,
                 wt  = if(ret.wt) cov$wt,
+                sqdist = if(ret.sqdist) res$sqdist,
+                prob= if(ret.pr) res$prob,
                 tol = tol,
                 eps = max(res$sqdist) - p,
-                it  = res$maxit)
+                it  = res$maxit,
+                maxit= maxit,
+                ierr = res$ierr,
+                conv = conv)
 
     class(res) <- "ellipsoid"
     res
@@ -65,6 +74,11 @@ print.ellipsoid <- function(x, digits = max(1, getOption("digits") - 2), ...)
     print(x$cov, digits = digits, ...)
     cat("  hence,",if(d==2)"area" else "volume"," = ",
         format(volume(x), digits=digits),"\n")
+    if(!x$conv) {
+        cat("\n** Warning: ** the algorithm did not terminate reliably!\n  ",
+            if(x$ierr) "most probably because of collinear data"
+            else "(in the available number of iterations)", "\n")
+    }
     invisible(x)
 }
 
