@@ -1,3 +1,4 @@
+/* -*- mode: c; kept-new-versions: 25; kept-old-versions: 20 -*- */
 
 /*   Clustering LARge Applications
      ~		~~~   ~
@@ -14,6 +15,7 @@
 #include <math.h>
 
 #include <R_ext/Print.h>/* for diagnostics */
+#include <R_ext/Random.h>/* when R's RNG is used */
 
 #include "cluster.h"
 
@@ -30,7 +32,9 @@ void clara(int *n,  /* = number of objects */
 	   int *mdata,	/*= {0,1}; 1: min(x) is missing value (NA);  0: no NA */
 	   double *valmd,/*[j]= missing value code (instead of NA) for x[,j]*/
 	   int *jtmd,	/* [j]= {-1,1};	 -1: x[,j] has NA; 1: no NAs in x[,j] */
-	   int *diss_kind,/* = {1,2};  1 : euclidean;  2 : manhattan*/
+	   int *diss_kind,/*= {1,2};  1 : euclidean;  2 : manhattan*/
+	   int *rng_R,	/*  = {0,1};  0 : use clara's internal weak RNG;
+			 *	      1 : use R's RNG (and seed) */
 	   int *nrepr,
 	   int *nsel,
 	   int *nbest,/* x[nbest[j]] will be the j-th obs in the final sample */
@@ -57,32 +61,30 @@ void clara(int *n,  /* = number of objects */
 
     /* Local variables */
 
-    Rboolean nafs, kall, full_sample, has_NA = *mdata;
-    int j, jk, jkk, js, jsm, jhalt, jran, l, less;
+    Rboolean nafs, kall, full_sample, lrg_sam, has_NA = *mdata;
+    int j, jk, jkk, js, jsm, jhalt, jran, l, n_sam;
     int nsm, ntt, nad, nadv, kran, kans, nrun, nexap, nexbp;
     int n_dys, nsamb, nunfs;
-    double rnn, sky, zb, zba = -1., s = -1., sx = -1.;/* Wall */
-
-    /* Parameter adjustments */
-    --nsel;
+    double rnn, sky, zb, s, sx = -1., zba = -1.;/* Wall */
 
     *jstop = 0;
     rnn = (double) (*n);
     /* n_dys := size of distance array dys[] */
     n_dys = *nsam * (*nsam - 1) / 2 + 1;/* >= 1 */
-
     full_sample = (*n == *nsam);/* only one sub sample == full data */
-
-    nsamb = *nsam << 1;
-    if (*n < nsamb)/* sample more than *n/2 -->
-		    * generate indices for smaller half */
-	less = *n - *nsam;
+    nsamb = *nsam * 2;
+    lrg_sam = (*n < nsamb);/* sample more than *n/2 */
+    if (lrg_sam)/* generate indices for the other, smaller half */
+	n_sam = *n - *nsam;
     else
-	less = *nsam;
+	n_sam = *nsam;
     nunfs = 0;
     kall = FALSE;
 
-    nrun = 0; /* << initialize `random seed' of the very simple randm() below */
+    if(*rng_R)
+	GetRNGstate();
+    else /* << initialize `random seed' of the very simple randm() below */
+	nrun = 0;
 
 /* __LOOP__ :  random subsamples are drawn and partitioned into kk clusters */
 
@@ -90,14 +92,15 @@ void clara(int *n,  /* = number of objects */
 	jhalt = 0;
 	if (!full_sample) {/* `real' case: sample size < n */
 	    ntt = 0;
-	    if (kall/*was jran != 1 */ && nunfs != jran && *n >= nsamb) {
+	    if (kall/*was jran != 1 */ && nunfs != jran && !lrg_sam) {
 		/* nsel[] := sort(nrx[])   for the first j=1:k	?? */
 		for (jk = 0; jk < *kk; ++jk)
-		    nsel[jk+1] = nrx[jk];
-		for (jk = 1; jk < *kk; ++jk) {
+		    nsel[jk] = nrx[jk];
+		for (jk = 0; jk < *kk-1; ++jk) {
+		    /* FIXME: nsel[] is 0-indexed, but *contains* 1-indices*/
 		    nsm = nsel[jk];
 		    jsm = jk;
-		    for (jkk = jk + 1; jkk <= *kk; ++jkk) {
+		    for (jkk = jk + 1; jkk < *kk; ++jkk) {
 			if (nsel[jkk] < nsm) {
 			    nsm = nsel[jkk];
 			    jsm = jkk;
@@ -109,79 +112,96 @@ void clara(int *n,  /* = number of objects */
 		ntt = *kk;
 	    }
 	    else {
-
-		/* Loop find random index `kran' not yet in nrx[] : */
+		/* Loop finding random index `kran' not yet in nrx[] : */
 	    L180:
-		kran = (int) (rnn * randm(&nrun) + 1.);
-		if (kran > *n) kran = *n;
-
+		kran = 1 + (int)(rnn * ((*rng_R)? unif_rand(): randm(&nrun)));
+		if(*trace_lev >= 3) {
+		    if(*rng_R)
+			Rprintf("... R unif_rand() -> k{ran}=%d\n", kran);
+		    else
+			Rprintf("... {180} nrun=%d -> k{ran}=%d\n", nrun,kran);
+		}
 		if (kall/*jran != 1*/) {
-		    for (jk = 0; jk < *kk; ++jk) {
+		    for (jk = 0; jk < *kk; ++jk)
 			if (kran == nrx[jk])
 			    goto L180;
-		    }
 		}
 		/* end Loop */
-		++ntt;
 		nsel[ntt] = kran;
-		if (ntt == less)
-		    goto L290;
+		if (++ntt == n_sam)
+		    goto L295;
+	    }
+
+	    if(*trace_lev >= 2) {
+		Rprintf(".. kall(T/F)=%d , nsel[ntt=%d] = %d\n",
+			kall, ntt, nsel[ntt]);
+		if(*trace_lev >= 3) {
+		    Rprintf("... nrx[0:%d]= ",*kk-1);
+		    for (jk = 0; jk < *kk; jk++)
+			Rprintf("%d ",nrx[jk]); Rprintf("\n");
+		    Rprintf("... nsel[1:%d]= ",ntt);
+		    for (jk = 1; jk <= ntt; jk++)
+			Rprintf("%d ",nsel[jk]); Rprintf("\n");
+		}
 	    }
 
 	    do {
 	    L210:
+		/* find `kran', a random `k' in {1:n},
+		 * not in nrx[0:(k-1)] nor nsel[1:ntt] : */
 		kran = (int) (rnn * randm(&nrun) + 1.);
+		if(*trace_lev >= 3)
+		    Rprintf("... {210} nrun=%d -> k{ran}=%d\n", nrun,kran);
 		if (kran > *n) kran = *n;
 
-		if (kall/*jran != 1*/ && *n < nsamb) {
+		if (kall/*jran != 1*/ && lrg_sam) {
 		    for (jk = 0; jk < *kk; ++jk) {
 			if (kran == nrx[jk])
 			    goto L210;
 		    }
 		}
-		/* insert kran into nsel[] :   (?) */
-		for (kans = 1; kans <= ntt; ++kans) {
+		/* insert kran into nsel[1:ntt] or after  and increase ntt : */
+		for (kans = 0; kans < ntt; ++kans)
 		    if (nsel[kans] >= kran) {
 			if (nsel[kans] == kran)
 			    goto L210;
 			else {
 			    for (nad = kans; nad <= ntt; ++nad) {
 				nadv = ntt - nad + kans;
-				nsel[nadv + 1] = nsel[nadv];
+				nsel[nadv] = nsel[nadv-1];
 			    }
-			    ++ntt;
 			    nsel[kans] = kran;
-			    goto L290;
+			    /* continue _outer_ loop */ goto L290;
 			}
 		    }
-		}
-		++ntt;
 		nsel[ntt] = kran;
-	    L290:
-		;
-	    } while (ntt < less);
 
-	    if (*n < nsamb) {
+	    L290:
+		++ntt;
+	    } while (ntt < n_sam);
+
+	L295:
+	    if (lrg_sam) {
 		/* have indices for smaller _nonsampled_ half; revert this: */
-		for (j = 1, nexap = 1, nexbp = 0; j < *n; j++) {
+		for (j = 1, nexap = 0, nexbp = 0; j <= *n; j++) {
 		    if (nsel[nexap] == j)
 			++nexap;
 		    else
 			nrepr[nexbp++] = j;
 		}
 		for (j = 0; j < *nsam; ++j)
-		    nsel[j+1] = nrepr[j];
+		    nsel[j] = nrepr[j];
 	    }
 	    if(*trace_lev)
 		Rprintf("C clara(): sample %d [ntt=%d, nunfs=%d] -> dysta2()\n",
 			jran, ntt, nunfs);
 	}
 	else { /* full_sample : *n = *nsam -- one sample is enough ! */
-	    for (j = 1; j <= *nsam; ++j)
-		nsel[j] = j;
+	    for (j = 0; j < *nsam; ++j)
+		nsel[j] = j+1;/* <- still uses 1-indices for its *values*! */
 	}
 
-	dysta2(*nsam, *jpp, &nsel[1], x, *n, dys, *diss_kind,
+	dysta2(*nsam, *jpp, nsel, x, *n, dys, *diss_kind,
 	       jtmd, valmd, &jhalt);
 	if (jhalt == 1) {
 	    if(*trace_lev)
@@ -197,12 +217,15 @@ void clara(int *n,  /* = number of objects */
 		s = dys[l];
 	} while (l+1 < n_dys);
 
+	if(*trace_lev >= 2)
+	    Rprintf(". clara(): s:= max dys[l=%d] = %g\n", l,s);
+
 	bswap2(*kk, *nsam, nrepr, dys, &sky, s,
 	       /* dysma */tmp1, /*dysmb*/tmp2,
 	       /* beter[], only used here */&tmp[nsamb]);
 
 	selec(*kk, *n, *jpp, *diss_kind, &zb, *nsam, has_NA, jtmd, valmd,
-	      nrepr, &nsel[1], dys, x, nr, &nafs, ttd, radus, ratt,
+	      nrepr, nsel, dys, x, nr, &nafs, ttd, radus, ratt,
 	      ntmp1, ntmp2, ntmp3, ntmp4, ntmp5, ntmp6, tmp1, tmp2);
 
 	if (nafs) {
@@ -218,7 +241,7 @@ void clara(int *n,  /* = number of objects */
 		nrx  [jk] = nr	 [jk];
 	    }
 	    for (js = 0; js < *nsam; ++js)
-		nbest[js] = nsel[js+1];
+		nbest[js] = nsel[js];
 	    sx = s;
 	}
 
@@ -227,16 +250,20 @@ void clara(int *n,  /* = number of objects */
 	if(full_sample) break; /* out of resampling */
     }
 /* --- end random sampling loop */
+    if(*rng_R)
+	PutRNGstate();
 
     if (nunfs >= *nran) { *jstop = 1; return; }
 
+
 /*     for the best subsample, the objects of the entire data set
      are assigned to their clusters */
 
     if (!kall) { *jstop = 2; return; }
 
-    if(*trace_lev)
-	Rprintf("C clara(): sample found. --> dysta2(nbest), resul(), finis\n");
+    if(*trace_lev) {
+	Rprintf("C clara(): sample _found_ --> dysta2(nbest), resul(), end\n");
+    }
     *obj = zba / rnn;
     dysta2(*nsam, *jpp, nbest, x, *n, dys, *diss_kind, jtmd, valmd, &jhalt);
 
@@ -275,17 +302,20 @@ void dysta2(int nsam, int jpp, int *nsel,
     for (l = 1; l < nsam; ++l) {
 	lsel = nsel[l];
 	if(lsel <= 0 || lsel > n)
-	    REprintf("	** dysta2(): nsel[l= %d] = %d is OUT\n", l, lsel);
+	    REprintf(" ** dysta2(): nsel[l= %d] = %d is OUT\n", l, lsel);
 	for (k = 0; k < l; ++k) {
 	    ksel = nsel[k];
+	    if(ksel <= 0 || ksel > n)
+		REprintf(" ** dysta2(): nsel[k= %d] = %d is OUT\n", k, ksel);
 	    clk = 0.;
 	    ++nlk;
 	    npres = 0;
 	    for (j = 0; j < jpp; ++j) {
-		lj = (lsel - 1) * jpp + j;
-		kj = (ksel - 1) * jpp + j;
+		lj = lsel-1 + j * n;
+		kj = ksel-1 + j * n;
 		if (jtmd[j] < 0) {
-/* in the following line, x[-2] ==> seg.fault {BDR to R-core, Sat, 3 Aug 2002} */
+		    /* in the following line (Fortran!), x[-2] ==> seg.fault
+		       {BDR to R-core, Sat, 3 Aug 2002} */
 		    if (x[lj] == valmd[j]) {
 			continue /* next j */;
 		    }
@@ -315,12 +345,12 @@ double randm(int *nrun)
 {
 /* we programmed this generator ourselves because we wanted it
    to be machine independent. it should run on most computers
-   because the largest int used is less than 2**30 . the period
-   is 2**16=65536, which is good enough for our purposes. */
+   because the largest int used is less than 2^30 . the period
+   is 2^16=65536, which is good enough for our purposes. */
     /* MM: improved the original speed-wise only: */
     *nrun = (*nrun * 5761 + 999) & 0177777;
     /* Masking off all but the last 16 bits is equivalent to  % 65536 */
-    return ((double) (*nrun) / 65536.f);
+    return ((double) (*nrun) / 65536.);
 } /* randm() */
 
 /* bswap2() : called once [per random sample] from clara() : */
@@ -468,7 +498,8 @@ void selec(int kk, int n, int jpp, int diss_kind,
 
     --dys;
 
-    /* Function Body */
+
+    /* nafs := TRUE  if a distance cannot be calculated*/
     *nafs = FALSE;
 
 /* identification of representative objects, and initializations */
@@ -500,8 +531,8 @@ void selec(int kk, int n, int jpp, int diss_kind,
 		nrjk = nr[jk];
 		if (nrjk != jj) {
 		    for (jp = 0; jp < jpp; ++jp) {
-			na = (nrjk - 1) * jpp + jp;
-			nb = (jj   - 1) * jpp + jp;
+			na = (nrjk - 1) + jp * n;
+			nb = (jj   - 1) + jp * n;
 			tra = fabs(x[na] - x[nb]);
 			if (diss_kind == 1)
 			    tra *= tra;
@@ -523,8 +554,8 @@ void selec(int kk, int n, int jpp, int diss_kind,
 		if (nrjk != jj) {
 		    nobs = 0;
 		    for (jp = 0; jp < jpp; ++jp) {
-			na = (nrjk - 1) * jpp + jp;
-			nb = (jj   - 1) * jpp + jp;
+			na = (nrjk - 1) + jp * n;
+			nb = (jj   - 1) + jp * n;
 			if (jtmd[jp] < 0) {
 			    if (x[na] == valmd[jp] || x[nb] == valmd[jp])
 				continue /* next jp */;
@@ -650,16 +681,14 @@ void resul(int kk, int n, int jpp, int diss_kind, Rboolean has_NA,
 	    if (nrx[jk] == jj + 1)/* 1-indexing */
 		goto L220; /* continue next jj (i.e., outer loop) */
 	}
-	njnb = jj * jpp;
+	njnb = jj;
 
 	if (!has_NA) {
 	    for (jk = 0; jk < kk; ++jk) {
 		dsum = 0.;
-		nrjk = (nrx[jk] - 1) * jpp;
+		nrjk = (nrx[jk] - 1);
 		for (j = 0; j < jpp; ++j) {
-		    na = nrjk + j;
-		    nb = njnb + j;
-		    tra = fabs(x[na] - x[nb]);
+		    tra = fabs(x[nrjk + j * n] - x[njnb + j * n]);
 		    if (diss_kind == 1)
 			tra *= tra;
 		    dsum += tra;
@@ -677,11 +706,11 @@ void resul(int kk, int n, int jpp, int diss_kind, Rboolean has_NA,
 	else { /* _has_ missing data */
 	    for (jk = 0; jk < kk; ++jk) {
 		dsum = 0.;
-		nrjk = (nrx[jk] - 1) * jpp;
+		nrjk = (nrx[jk] - 1);
 		abc = 0.;
 		for (j = 0; j < jpp; ++j) {
-		    na = nrjk + j;
-		    nb = njnb + j;
+		    na = nrjk + j * n;
+		    nb = njnb + j * n;
 		    if (jtmd[j] < 0) {
 			if (x[na] == valmd[j] || x[nb] == valmd[j])
 			    continue /* next j */;
@@ -708,16 +737,17 @@ void resul(int kk, int n, int jpp, int diss_kind, Rboolean has_NA,
 
     L220:
 	;
-    } /* for(jj)  while (jj < n);*/
+    } /* for(jj = 0; jj < n ..)*/
 
-    for (jk = 0; jk < kk; ++jk) {
-	nrjk = nrx[jk];
-	x[(nrjk - 1) * jpp] = (double) jk + 1;/* 1-indexing */
-    }
+
+    for (jk = 0; jk < kk; ++jk)
+	x[nrx[jk] - 1] = (double) jk + 1;/* 1-indexing */
+
+    /* mtt[k] := size(k-th cluster) : */
     for (ka = 0; ka < kk; ++ka) {
 	mtt[ka] = 0;
 	for(j = 0; j < n; j++) {
-	    if (((int) (x[j * jpp] + .1f)) == ka + 1)/* 1-indexing */
+	    if (((int) x[j]) == ka + 1)/* 1-indexing */
 		++mtt[ka];
 	}
     }
@@ -766,7 +796,7 @@ void black(int kk, int jpp, int nsam, int *nbest,
     /* Function Body */
     for (l = 1; l <= nsam; ++l) {
 	ncase = nbest[l];
-	ncluv[l] = (int) (x[(ncase - 1) * jpp] + .1f);
+	ncluv[l] = (int) x[ncase - 1];
     }
 
 /*     drawing of the silhouettes */
